/*! Copyright (C) [2011] [David Mainzer]
 * 
 *  This program is free software; you can redistribute it and/or modify it under the terms 
 *  of the GNU General Public License as published by the Free Software Foundation; either 
 *  version 3 of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 *  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU General Public License for more details.
 *  
 *	You should have received a copy of the GNU General Public License along with this program; 
 *	if not, see <http://www.gnu.org/licenses/>.  
 */

import java.util.Iterator;
import java.util.Vector;
import java.util.HashMap;

import vrml.Event;
import vrml.field.MFInt32;
import vrml.field.MFVec3f;
import vrml.field.SFNode;
import vrml.node.Node;
import vrml.node.Script;

public class SpringMassSystem extends vrml.node.Script
{
//	field SFNode objectPlane USE coord_Plane
    private SFNode m_planePoints;
//	field SFNode objectPlane2 USE coord_Geo
    private SFNode m_planeIndex;

    // list of mass-points (index-IDs)
    private Vector<Spring>   m_Feder         = new Vector<Spring>();
    
    private Vector<MassPoint> m_massPoints = new Vector<MassPoint>();
    private HashMap<Integer,Integer> m_massPointsHash = new HashMap<Integer,Integer>();
    
    long runtime   = (long) 0.0;										//!< simulation run time (in ms)
    long lastruntime = (long) 0.0;										//!< Zeit seit letztem frame
    long starttime = (long) 0.0;										//!< simulation start time (in ms)
    
    
    /*! \brief initialization of the hole mass-spring system
    *  creates the springs
    * 
    */
    @Override
    public void initialize()
    {
        System.out.println("################################");
        System.out.println("## Spring-Mass-Sys using JAVA ##");
        System.out.println("################################");
        System.out.println("");
        System.out.println("Initializing java ... \n");
        
        // get the coordinates    	
        m_planePoints = (SFNode)getField("objectPlane");
        MFVec3f cloth_Points = (MFVec3f)((Node)m_planePoints.getValue()).getExposedField("point");

        // get the index coordinates
        m_planeIndex = (SFNode)getField("objectGeo");
        MFInt32 cloth_Points_Index = (MFInt32)((Node)m_planeIndex.getValue()).getExposedField("coordIndex");

        Vector<Vector<Integer>> federList = new Vector<Vector<Integer>>();
        federList.clear();


        if ( (cloth_Points_Index.getSize() % 5) != 0)
        {
            System.err.println("number of masspoints: " + cloth_Points_Index.getSize());
            System.err.println("we only support quad meshes");
            System.exit(0);
        }

        // creating mass-points and connections between mass-points
        for ( int i=0; i < cloth_Points_Index.getSize(); i=i+5 )
        {
            // get 4 indices
            final int tmp_Index_0 = cloth_Points_Index.get1Value(i + 0);
            final int tmp_Index_1 = cloth_Points_Index.get1Value(i + 1);
            final int tmp_Index_2 = cloth_Points_Index.get1Value(i + 2);
            final int tmp_Index_3 = cloth_Points_Index.get1Value(i + 3);

            final int tmp_Index_T = cloth_Points_Index.get1Value(i + 4);

            if ( tmp_Index_T != -1 )
            {
                System.err.println("we only support quad meshes");
                System.exit(0);
            }
            // t0 -- t1
            //  | \ / |
            //  | / \ |
            // t3 -- t2

            final float tmp_point_0 [] = new float[3];
            cloth_Points.get1Value(tmp_Index_0, tmp_point_0);
            
            final float tmp_point_1 [] = new float[3];
            cloth_Points.get1Value(tmp_Index_1, tmp_point_1);
            
            final float tmp_point_2 [] = new float[3];
            cloth_Points.get1Value(tmp_Index_2, tmp_point_2);
            
            final float tmp_point_3 [] = new float[3];
            cloth_Points.get1Value(tmp_Index_3, tmp_point_3);


            // massPoint with Idx tmp_Index_0 is not in the hash map
            if( ! m_massPointsHash.containsKey(tmp_Index_0) )
            {
                MassPoint tmp_massPoint = new MassPoint( tmp_Index_0, tmp_point_0 );
                // not all masspoints are movable
                if ( tmp_Index_0 != 228 && tmp_Index_0 != 132 && tmp_Index_0 != 231 && tmp_Index_0 != 168 &&
                     tmp_Index_0 != 232 && tmp_Index_0 != 118 && tmp_Index_0 != 243 && tmp_Index_0 != 180 && 
                     tmp_Index_0 != 244 && tmp_Index_0 != 140 && tmp_Index_0 != 130 && tmp_Index_0 !=  21 &&
                     tmp_Index_0 != 134 && tmp_Index_0 !=  51 && tmp_Index_0 != 133 && tmp_Index_0 !=  18 &&
                     tmp_Index_0 != 137 && tmp_Index_0 !=  53 && tmp_Index_0 != 138 && tmp_Index_0 !=  22    )
                {
                    tmp_massPoint.setMove(true);
                }
                else 
                {
                    tmp_massPoint.setMove(false);
                }
                    
                m_massPoints.add( tmp_massPoint );
                m_massPointsHash.put(tmp_Index_0, m_massPoints.indexOf(tmp_massPoint));
            }

            if( ! m_massPointsHash.containsKey(tmp_Index_1) )
            {
                MassPoint tmp_massPoint = new MassPoint( tmp_Index_1, tmp_point_1 );
                // not all masspoints are movable
                if ( tmp_Index_1 != 228 && tmp_Index_1 != 132 && tmp_Index_1 != 231 && tmp_Index_1 != 168 &&
                     tmp_Index_1 != 232 && tmp_Index_1 != 118 && tmp_Index_1 != 243 && tmp_Index_1 != 180 && 
                     tmp_Index_1 != 244 && tmp_Index_1 != 140 && tmp_Index_1 != 130 && tmp_Index_1 !=  21 &&
                     tmp_Index_1 != 134 && tmp_Index_1 !=  51 && tmp_Index_1 != 133 && tmp_Index_1 !=  18 &&
                     tmp_Index_1 != 137 && tmp_Index_1 !=  53 && tmp_Index_1 != 138 && tmp_Index_1 !=  22    )
                {
                    tmp_massPoint.setMove(true);
                } 
                else 
                {
                    tmp_massPoint.setMove(false);
                }
            
                m_massPoints.add( tmp_massPoint );
                m_massPointsHash.put(tmp_Index_1, m_massPoints.indexOf(tmp_massPoint));
            }

            if( ! m_massPointsHash.containsKey(tmp_Index_2) )
            {
                MassPoint tmp_massPoint = new MassPoint( tmp_Index_2, tmp_point_2 );
                // not all masspoints are movable
                if ( tmp_Index_2 != 228 && tmp_Index_2 != 132 && tmp_Index_2 != 231 && tmp_Index_2 != 168 &&
                     tmp_Index_2 != 232 && tmp_Index_2 != 118 && tmp_Index_2 != 243 && tmp_Index_2 != 180 && 
                     tmp_Index_2 != 244 && tmp_Index_2 != 140 && tmp_Index_2 != 130 && tmp_Index_2 !=  21 &&
                     tmp_Index_2 != 134 && tmp_Index_2 !=  51 && tmp_Index_2 != 133 && tmp_Index_2 !=  18 &&
                     tmp_Index_2 != 137 && tmp_Index_2 !=  53 && tmp_Index_2 != 138 && tmp_Index_2 !=  22    )
                {
                    tmp_massPoint.setMove(true);
                } 
                else 
                {
                    tmp_massPoint.setMove(false);
                }
                
                m_massPoints.add( tmp_massPoint );
                m_massPointsHash.put(tmp_Index_2, m_massPoints.indexOf(tmp_massPoint));
            }
            
            if( ! m_massPointsHash.containsKey(tmp_Index_3) )
            {
                MassPoint tmp_massPoint = new MassPoint( tmp_Index_3, tmp_point_3 );
                // not all masspoints are movable
                if ( tmp_Index_3 != 228 && tmp_Index_3 != 132 && tmp_Index_3 != 231 && tmp_Index_3 != 168 &&
                     tmp_Index_3 != 232 && tmp_Index_3 != 118 && tmp_Index_3 != 243 && tmp_Index_3 != 180 && 
                     tmp_Index_3 != 244 && tmp_Index_3 != 140 && tmp_Index_3 != 130 && tmp_Index_3 !=  21 &&
                     tmp_Index_3 != 134 && tmp_Index_3 !=  51 && tmp_Index_3 != 133 && tmp_Index_3 !=  18 &&
                     tmp_Index_3 != 137 && tmp_Index_3 !=  53 && tmp_Index_3 != 138 && tmp_Index_3 !=  22    )
                {
                    tmp_massPoint.setMove(true);
                } 
                else
                {
                    tmp_massPoint.setMove(false);
                }

                m_massPoints.add( tmp_massPoint );
                m_massPointsHash.put(tmp_Index_3, m_massPoints.indexOf(tmp_massPoint));
            }

            // Spring 0 --> 1
            Vector<Integer> spring_01 = new Vector<Integer>();
            Vector<Integer> spring_10 = new Vector<Integer>();

            spring_01.add(tmp_Index_0);
            spring_01.add(tmp_Index_1);

            spring_10.add(tmp_Index_1);
            spring_10.add(tmp_Index_0);

            if( ! federList.contains(spring_01) && ! federList.contains(spring_10) )
            {
                federList.add(spring_01);

                Vector3 feder0_pos = new Vector3(tmp_point_0[0], tmp_point_0[1], tmp_point_0[2]);
                Vector3 feder1_pos = new Vector3(tmp_point_1[0], tmp_point_1[1], tmp_point_1[2]);

                Vector3 spring = feder0_pos.sub(feder1_pos);

                Spring tmp_Feder = new Spring(m_massPointsHash.get(tmp_Index_0),  m_massPointsHash.get(tmp_Index_1), spring.norm());

                // System.out.println("add Spring: " + spring_01.get(0) + " ---- " + spring_01.get(1) + "  ||  " + tmp_Feder.getId1() + " " + tmp_Feder.getId2() );

                m_Feder.add(tmp_Feder);
            }

            // Spring 0 --> 2
            Vector<Integer> spring_02 = new Vector<Integer>();
            Vector<Integer> spring_20 = new Vector<Integer>();
            
            spring_02.add(tmp_Index_0);
            spring_02.add(tmp_Index_2);
            
            spring_20.add(tmp_Index_2);
            spring_20.add(tmp_Index_0);
            
            if( ! federList.contains(spring_02) && ! federList.contains(spring_20) )
            {
                federList.add(spring_02);

                Vector3 feder0_pos = new Vector3(tmp_point_0[0], tmp_point_0[1], tmp_point_0[2] );
                Vector3 feder2_pos = new Vector3(tmp_point_2[0], tmp_point_2[1], tmp_point_2[2]);

                Vector3 spring = feder0_pos.sub(feder2_pos);

                Spring tmp_Feder = new Spring(m_massPointsHash.get(tmp_Index_0),  m_massPointsHash.get(tmp_Index_2), spring.norm());

                // System.out.println("add Spring: " + spring_02.get(0) + " ---- " + spring_02.get(1) + "  ||  " + tmp_Feder.getId1() + " " + tmp_Feder.getId2() );

                m_Feder.add(tmp_Feder);
            }

            // Spring 0 --> 3
            Vector<Integer> spring_03 = new Vector<Integer>();
            Vector<Integer> spring_30 = new Vector<Integer>();
            
            spring_03.add(tmp_Index_0);
            spring_03.add(tmp_Index_3);
            
            spring_30.add(tmp_Index_3);
            spring_30.add(tmp_Index_0);
            
            if( ! federList.contains(spring_03) && ! federList.contains(spring_30) )
            {
                federList.add(spring_03);

                Vector3 feder0_pos = new Vector3( tmp_point_0[0], tmp_point_0[1], tmp_point_0[2] );
                Vector3 feder3_pos = new Vector3( tmp_point_3[0], tmp_point_3[1], tmp_point_3[2] );
                    
                Vector3 spring = feder0_pos.sub(feder3_pos);
        
                Spring tmp_Feder = new Spring(m_massPointsHash.get(tmp_Index_0),  m_massPointsHash.get(tmp_Index_3), spring.norm());

                // System.out.println("add Spring: " + spring_03.get(0) + " ---- " + spring_03.get(1) + "  ||  " + tmp_Feder.getId1() + " " + tmp_Feder.getId2() );

                m_Feder.add(tmp_Feder);
            }

    		// Spring 1 --> 2
    		Vector<Integer> spring_12 = new Vector<Integer>();
    		Vector<Integer> spring_21 = new Vector<Integer>();
    		
    		spring_12.add(tmp_Index_1);
    		spring_12.add(tmp_Index_2);
    		
    		spring_21.add(tmp_Index_2);
    		spring_21.add(tmp_Index_1);
    		
    		if( ! federList.contains(spring_12) && ! federList.contains(spring_21) )
    		{
    			federList.add(spring_12);

    			Vector3 feder1_pos = new Vector3(tmp_point_1[0], tmp_point_1[1], tmp_point_1[2]);
    			Vector3 feder2_pos = new Vector3(tmp_point_2[0], tmp_point_2[1], tmp_point_2[2]);
    			
    			Vector3 spring = feder1_pos.sub(feder2_pos);
    			
    			Spring tmp_Feder = new Spring(m_massPointsHash.get(tmp_Index_1),  m_massPointsHash.get(tmp_Index_2), spring.norm());
    			
    			// System.out.println("add Spring: " + spring_12.get(0) + " ---- " + spring_12.get(1) + "  ||  " + tmp_Feder.getId1() + " " + tmp_Feder.getId2() );
    			
    			m_Feder.add(tmp_Feder);
    		}
    		
    		// Spring 1 --> 3
    		Vector<Integer> spring_13 = new Vector<Integer>();
    		Vector<Integer> spring_31 = new Vector<Integer>();
    		
    		spring_13.add(tmp_Index_1);
    		spring_13.add(tmp_Index_3);
    		
    		spring_31.add(tmp_Index_3);
    		spring_31.add(tmp_Index_1);
    		
    		if( ! federList.contains(spring_13) && ! federList.contains(spring_31) )
    		{
    			federList.add(spring_13);

    			Vector3 feder1_pos = new Vector3(tmp_point_1[0], tmp_point_1[1], tmp_point_1[2]);
    			Vector3 feder3_pos = new Vector3(tmp_point_3[0], tmp_point_3[1], tmp_point_3[2]);
    			
    			Vector3 spring = feder1_pos.sub(feder3_pos);
    			
    			Spring tmp_Feder = new Spring(m_massPointsHash.get(tmp_Index_1),  m_massPointsHash.get(tmp_Index_3), spring.norm());
    			
    			// System.out.println("add Spring: " + spring_13.get(0) + " ---- " + spring_13.get(1) + "  ||  " + tmp_Feder.getId1() + " " + tmp_Feder.getId2() );
    			
    			m_Feder.add(tmp_Feder);
    		}
    		
    		// Spring 2 --> 3
    		Vector<Integer> spring_23 = new Vector<Integer>();
    		Vector<Integer> spring_32 = new Vector<Integer>();
    		
    		spring_23.add(tmp_Index_2);
    		spring_23.add(tmp_Index_3);
    		
    		spring_32.add(tmp_Index_3);
    		spring_32.add(tmp_Index_2);
    		
    		if( ! federList.contains(spring_23) && ! federList.contains(spring_32) )
    		{
    			federList.add(spring_23);

    			Vector3 feder2_pos = new Vector3(tmp_point_2[0], tmp_point_2[1], tmp_point_2[2]);
    			Vector3 feder3_pos = new Vector3(tmp_point_3[0], tmp_point_3[1], tmp_point_3[2]);
    			
    			Vector3 spring = feder2_pos.sub(feder3_pos);
    			
    			Spring tmp_Feder = new Spring(m_massPointsHash.get(tmp_Index_2),  m_massPointsHash.get(tmp_Index_3), (spring.norm()) );
    			
    			// System.out.println("add Spring: " + spring_23.get(0) + " ---- " + spring_23.get(1) + "  ||  " + tmp_Feder.getId1() + " " + tmp_Feder.getId2() );
    			
    			m_Feder.add(tmp_Feder);
    		}
    	}
    	
    	System.out.println("Number of mass points: " + m_massPoints.size() );
    	System.out.println("Number of springs: " + m_Feder.size() );
    	// set start time
		starttime = System.currentTimeMillis();
	}

	/** set uptime of the simulation (in ms)
	 * 
	 */
	public void setTimer ()
	{
		long timeInMillis = System.currentTimeMillis();
		long newruntime = timeInMillis - starttime;
		lastruntime = newruntime - runtime;
		runtime = newruntime;
	}

	@Override
    public void processEvent(Event e) 
	{
		setTimer();
		simulate();
	}

	public void simulate()
	{
		// get the coordinates    	
		MFVec3f cloth_Points = (MFVec3f)((Node)m_planePoints.getValue()).getExposedField("point");
    	
    	Iterator<Spring> it = m_Feder.iterator();
    	
		lastruntime = runtime;

    	while (it.hasNext()) {
    		Spring tmp_Spring = it.next();

    		int massPoint1PosVec = tmp_Spring.getId1();
    		int massPoint2PosVec = tmp_Spring.getId2();

    		if( m_massPoints.size() <= massPoint1PosVec || m_massPoints.size() <= massPoint2PosVec )
    		{
    			System.out.println("ERROR");
    		}
	    		
    		MassPoint massPoint1 = m_massPoints.get( massPoint1PosVec );
    		MassPoint massPoint2 = m_massPoints.get( massPoint2PosVec );
        		
    		// Teil a): bestimmen Sie die Krfte zwischen diesen zwei Masse-Punkten!
			// Teil b): 
    		
    		// so fgen Sie eine Kraft zu einem der Punktmassen hinzu
    		// massPoint1.addForce( 1.0, 1.0, 1.0 );
    	}
	    	
    	Iterator<MassPoint> itMassPoint = m_massPoints.iterator();
    	while (itMassPoint.hasNext()) {
    		MassPoint tmp_massPoint   = itMassPoint.next(); 
    		int       tmp_massPointID = tmp_massPoint.getID();

    		// hier wirkt die Gravitation auf die Punktmassen
//			Vector3 vGravity1 = new Vector3( 0.f, -tmp_massPoint.getMass()*9.81f, 0.f );
//			tmp_massPoint.addForce( vGravity1 );

			// anwenden der Krfte auf die Punktmassen
			tmp_massPoint.setOldPos();
    		tmp_massPoint.updateCurPos( );

    		// schreibe die neuen Positionen zurck in die Scene
    		if ( tmp_massPoint.getMove() )
    		{		    		
    			cloth_Points.set1Value(tmp_massPointID, tmp_massPoint.getX(), tmp_massPoint.getY(), tmp_massPoint.getZ());
    		}
	    }
	}
}


